package helm

import (
	"archive/tar"
	"bytes"
	"compress/gzip"
	"encoding/base64"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"reflect"
	"strconv"
	"strings"

	"dario.cat/mergo"
	"github.com/samber/lo"
	log "github.com/sirupsen/logrus"
	"gopkg.in/yaml.v3"
)

// GetValuesMapFromChart takes a path to a values file in an encoded helm chart and returns it values in a map format
func GetValuesMapFromChart(valuesFilePath, base64EncodedChart string) (valuesMap map[string]interface{}, err error) {
	chartArchive, err := base64.StdEncoding.DecodeString(base64EncodedChart)
	if err != nil {
		return nil, fmt.Errorf("error decoding base64-encoded chart archive: %v", err)
	}
	chart, err := loadChart(chartArchive)
	if err != nil {
		return nil, fmt.Errorf("error loading chart: %v", err)
	}

	if valuesFilePath == "values.yaml" {
		return chart.Values, nil
	}

	for _, f := range chart.Files {
		if f.Name == valuesFilePath {
			valuesMap = make(map[string]interface{})
			err = yaml.Unmarshal(f.Data, valuesMap)
			if err != nil {
				return nil, err
			}
			break
		}
	}
	if valuesMap == nil {
		return nil, NewErrorValuesFileNotFound(valuesFilePath)
	}
	return valuesMap, nil
}

func ProcessYAMLFiles(filePaths []string) (map[string]interface{}, error) {
	mergedMap := make(map[string]interface{})

	for _, filePath := range filePaths {
		fileContent, err := ioutil.ReadFile(filePath)
		if err != nil {
			return nil, err
		}

		var yamlMap map[string]interface{}
		err = yaml.Unmarshal(fileContent, &yamlMap)
		if err != nil {
			return nil, err
		}

		mergedMap = lo.Assign(mergedMap, yamlMap)
		if err != nil {
			return nil, err
		}
	}

	return mergedMap, nil
}

// applyDefaultEnvironmentBasedValues takes a encoded chart, passed values and applies the appropriate default config for an env,cloud,project etc
func applyDefaultEnvironmentBasedValues(chart string, values map[string]interface{}) (mergedValues map[string]interface{}, err error) {
	// Always apply default values.yaml
	mergedValues, err = MergeValuesFromChartPath(chart, "values.yaml", mergedValues)
	if err != nil {
		return nil, err
	}
	for _, checkMatch := range []struct {
		matchKey       string
		matchValue     []interface{}
		valuesFilePath string
	}{
		{"cloud", nil, "{value}/values.yaml"},
		{"controllerConfig.variant", nil, "{cloud}/variant/values-{value}.yaml"},
		{"project.tenantProjectNumber", nil, "{cloud}/project/values-{value}.yaml"},
		{"regional", []interface{}{true, "true"}, "{cloud}/type/values-regional.yaml"},
		{"featureFlags.enableGcp1p", []interface{}{true, "true"}, "{cloud}/type/values-gcp1p.yaml"},
	} {
		value, found, err := GetPathValue(checkMatch.matchKey, values)
		if err != nil {
			log.WithFields(
				log.Fields{
					"key":   checkMatch,
					"error": err,
				},
			).Debug("ignoring GetPathValue error")
			continue
		}
		if found {
			cloudVal, cloudValOk := values["cloud"].(string)
			// Check that cloud is set before we can replace it in the values.yaml path
			if cloudValOk {
				checkMatch.valuesFilePath = strings.ReplaceAll(checkMatch.valuesFilePath, "{cloud}", cloudVal)
			}
			// Check that the value of the match is a string before we can replace it in the values.yaml path
			switch v := value.(type) {
			case string:
				checkMatch.valuesFilePath = strings.ReplaceAll(checkMatch.valuesFilePath, "{value}", v)
			case int:
				checkMatch.valuesFilePath = strings.ReplaceAll(checkMatch.valuesFilePath, "{value}", strconv.Itoa(v))
			}
			// If matchValue is set and the key we are checking (e.g. regional)
			// does not match the expected value (e.g true) skip this iteration
			if checkMatch.matchValue != nil {
				found := false
				for _, allowedMatch := range checkMatch.matchValue {
					if reflect.DeepEqual(value, allowedMatch) {
						found = true
					}
				}
				if !found {
					log.WithFields(
						log.Fields{
							"key":         checkMatch.matchKey,
							"matchValue":  checkMatch.matchValue,
							"actualValue": value,
							"file":        checkMatch.valuesFilePath,
						}).Debug("Skipped applying file as match and actual are not equal")
					continue
				}
			}
			log.WithFields(log.Fields{
				"file": checkMatch.valuesFilePath,
			}).Debug("Using values.yaml override")
			mergedValues, err = MergeValuesFromChartPath(chart, checkMatch.valuesFilePath, mergedValues)
			if err != nil {
				if IsErrorValuesFileNotFound(err) {
					log.WithFields(
						log.Fields{
							"key":  checkMatch.matchKey,
							"file": checkMatch.valuesFilePath,
							"err":  err,
						}).Debug("Skipping missing environment values file ")
					continue
				}
				return nil, err
			}
		} else {
			log.WithField("matchKey", checkMatch.matchKey).Debug("Could not find key")
		}
	}
	// one final merge to place the override values on top of the values.yaml files
	mergedValues, err = MergeMaps(mergedValues, values)
	if err != nil {
		return nil, err
	}
	return mergedValues, nil
}

// MergeValuesFromChartPath takes a encoded helm chart, a path to a values.yaml file and a map of existing values
// and merges the (values inside the) values.yaml file into the existing values
func MergeValuesFromChartPath(chart, valuesFile string, prevValues map[string]interface{}) (mergedValues map[string]interface{}, err error) {
	chartval, err := GetValuesMapFromChart(valuesFile, chart)
	if err != nil {
		return prevValues, err
	}
	// apply the previous values to ones from the file, note that the order is important

	mergedValues, err = MergeMaps(prevValues, chartval)
	if err != nil {
		return prevValues, err
	}
	return mergedValues, nil
}

func MergeMaps(prev, new map[string]interface{}) (mergedMap map[string]interface{}, err error) {
	original := &prev
	err = mergo.Merge(original, new,
		mergo.WithOverrideEmptySlice,
		mergo.WithOverride,
	)
	if err != nil {
		return nil, err
	}
	return *original, nil
}

// SetPathValue takes a key=value pair and replaces its tree path(level1.level2) to "value"
func SetPathValue(pair string, m map[string]interface{}) error {
	parts := strings.Split(pair, "=")
	if len(parts) != 2 {
		return fmt.Errorf("invalid key=value pair")
	}

	keyPath := strings.Split(parts[0], ".")
	var value interface{}
	switch parts[1] {
	case "true":
		value = true
	case "false":
		value = false
	default:
		value = parts[1]
	}

	var currentValue interface{} = m
	for i, key := range keyPath {
		if i == len(keyPath)-1 {
			switch currentValue.(type) {
			case map[string]interface{}:
				currentValue.(map[string]interface{})[key] = value
			default:
				return fmt.Errorf("invalid path in map")
			}
		} else {
			switch currentValue.(type) {
			case map[string]interface{}:
				if currentValue.(map[string]interface{})[key] == nil {
					currentValue.(map[string]interface{})[key] = make(map[string]interface{})
				}
				currentValue = currentValue.(map[string]interface{})[key]
			default:
				return fmt.Errorf("invalid path in map")
			}
		}
	}
	return nil
}

// GetPathValue takes a key path (level1.level2.level3) and returns the value at that path in a map[string]interface{}
func GetPathValue(keyPath string, m map[string]interface{}) (value interface{}, found bool, err error) {
	keys := strings.Split(keyPath, ".")

	var currentValue interface{} = m
	for i, key := range keys {
		switch currentValue.(type) {
		case map[string]interface{}:
			if currentValue.(map[string]interface{})[key] == nil {
				return nil, false, fmt.Errorf("key not found in map")
			}
			currentValue = currentValue.(map[string]interface{})[key]
		default:
			if i == len(keys)-1 {
				return currentValue, true, nil
			}
			return nil, false, fmt.Errorf("invalid path in map")
		}
	}

	return currentValue, true, nil
}

// MergeHelmValuesAndSetParam combines a list of values.yaml files and a list of key=value paremeters (in that order)
func MergeHelmValuesAndSetParam(valuesFiles, sets []string) (values map[string]interface{}, err error) {
	if log.GetLevel() == log.DebugLevel {
		log.WithFields(
			log.Fields{
				"valueFiles": valuesFiles,
				"sets":       sets,
			},
		).Debug("MergeHelmValuesAndSetParam input")
	}
	quarkValues, err := ProcessYAMLFiles(valuesFiles)
	if err != nil {
		return nil, err
	}
	for _, v := range sets {
		err := SetPathValue(v, quarkValues)
		if err != nil {
			return nil, err
		}
	}
	return quarkValues, nil
}

// GenerateBase64EncodedTar takes a path to the root folder of a helm chart and returns it's base64 (tar+gzip) equivalent
func GenerateBase64EncodedTar(chartPath string) (encodedchart string, err error) {
	chartPath, err = filepath.Abs(chartPath)
	if err != nil {
		return "", err
	}
	var buf bytes.Buffer
	gz := gzip.NewWriter(&buf)
	tw := tar.NewWriter(gz)

	// recurse though the chart path
	err = filepath.Walk(chartPath, func(path string, info os.FileInfo, err error) error {

		if err != nil {
			return err
		}

		// grab the file info
		header, err := tar.FileInfoHeader(info, info.Name())
		if err != nil {
			return err
		}
		// fix macos issues with relative pathing
		header.Name = strings.ReplaceAll(path, chartPath, "")

		if err := tw.WriteHeader(header); err != nil {
			return err
		}

		// compress and write a file to the archive
		if !info.IsDir() {
			data, err := ioutil.ReadFile(path)
			if err != nil {
				return err
			}
			if _, err := tw.Write(data); err != nil {
				return err
			}
		}

		return nil
	})

	if err != nil {
		return "", fmt.Errorf("error creating tar archive: %v", err)
	}

	if err := tw.Close(); err != nil {
		return "", fmt.Errorf("error closing tar writer: %v", err)
	}

	if err := gz.Close(); err != nil {
		return "", fmt.Errorf("error closing gzip writer: %v", err)
	}

	return base64.StdEncoding.EncodeToString(buf.Bytes()), nil
}
